# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.filter import DatasetItemEncoder
from datumaro.components.project import ProjectBuildTargets
from datumaro.util import str_to_bool
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import FilterModes, load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Extract subdataset",
        description="""
        Extracts a subdataset that contains only items matching filter.|n
        |n
        By default, datasets are updated in-place. The '-o/--output-dir'
        option can be used to specify another output directory. When
        updating in-place, use the '--overwrite' parameter (in-place
        updates fail by default to prevent data loss), unless a project
        target is modified.|n
        |n
        A filter is an XPath expression, which is applied to XML
        representation of a dataset item. Check '--dry-run' parameter
        to see XML representations of the dataset items.|n
        |n
        To filter annotations use the mode ('-m') parameter.|n
        Supported modes:|n
        - 'i', 'items'|n
        - 'a', 'annotations'|n
        - 'i+a', 'a+i', 'items+annotations', 'annotations+items'|n
        When filtering annotations, use the 'items+annotations'
        mode to point that annotation-less dataset items should be
        removed. To select an annotation, write an XPath that
        returns 'annotation' elements (see examples).|n
        |n
        This command has the following invocation syntax:
        - %(prog)s <target dataset revpath>|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        The current project (-p/--project) is also used as a context for
        plugins, so it can be useful for dataset paths having custom formats.
        When not specified, the current project's working tree is used.|n
        |n
        The command can be applied to a dataset or a project build target,
        a stage or the combined 'project' target, in which case all the
        targets will be affected. A build tree stage will be recorded
        if '--stage' is enabled, and the resulting dataset(-s) will be
        saved if '--apply' is enabled.|n
        |n
        Examples:|n
        - Filter images with width < height:|n
        |s|s%(prog)s -e '/item[image/width < image/height]'|n
        |n
        - Filter images with large-area bboxes:|n
        |s|s%(prog)s -e '/item[annotation/type="bbox" and
            annotation/area>2000]'|n
        |n
        - Filter out all irrelevant annotations from items:|n
        |s|s%(prog)s -m a -e '/item/annotation[label = "person"]'|n
        |n
        - Filter out all irrelevant annotations from items:|n
        |s|s%(prog)s -m a -e '/item/annotation[label="cat" and
        area > 99.5]'|n
        |n
        - Filter occluded annotations and items, if no annotations left:|n
        |s|s%(prog)s -m i+a -e '/item/annotation[occluded="True"]'|n
        |n
        - Filter a VOC-like dataset inplace:|n
        |s|s%(prog)s -e '/item/annotation[label = "bus"]' --overwrite dataset/:voc
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target", nargs="?", default="project", help="Target dataset revpath (default: %(default)s)"
    )
    parser.add_argument("-e", "--filter", help="XML XPath filter expression for dataset items")
    parser.add_argument(
        "-m",
        "--mode",
        default=FilterModes.i.name,
        type=FilterModes.parse,
        help="Filter mode (options: %s; default: %s)"
        % (", ".join(FilterModes.list_options()), "%(default)s"),
    )
    parser.add_argument(
        "--dry-run", action="store_true", help="Print XML representations to be filtered and exit"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="""
            Output directory. Can be omitted for main project targets
            (i.e. data sources and the 'project' target, but not
            intermediate stages) and dataset targets.
            If not specified, the results will be saved inplace.
            """,
    )
    parser.add_argument(
        "--stage",
        type=str_to_bool,
        default=True,
        help="""
            Include this action as a project build step.
            If true, this operation will be saved in the project
            build tree, allowing to reproduce the resulting dataset later.
            Applicable only to main project targets (i.e. data sources
            and the 'project' target, but not intermediate stages)
            (default: %(default)s)
            """,
    )
    parser.add_argument(
        "--apply",
        type=str_to_bool,
        default=True,
        help="Run this command immediately. If disabled, only the "
        "build tree stage will be written (default: %(default)s)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=filter_command)

    return parser


def get_sensitive_args():
    return {
        filter_command: [
            "target",
            "filter",
            "dst_dir",
            "project_dir",
        ],
    }


@scoped
def filter_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    filter_args = FilterModes.make_filter_args(args.mode)
    filter_expr = args.filter

    if args.dry_run:
        dataset, _project = parse_full_revpath(args.target, project)
        if _project:
            scope_add(_project)

        dataset = dataset.filter(expr=filter_expr, **filter_args)

        for item in dataset:
            encoded_item = DatasetItemEncoder.encode(item, dataset.categories())
            xml_item = DatasetItemEncoder.to_string(encoded_item)
            print(xml_item)
        return 0

    if not args.filter:
        raise CliException("Expected a filter expression ('-e' argument)")

    is_target = project is not None and args.target in project.working_tree.build_targets
    if is_target:
        if (
            not args.dst_dir
            and args.stage
            and (args.target != ProjectBuildTargets.strip_target_name(args.target))
        ):
            raise CliException(
                "Adding a stage is only allowed for " "project targets, not their stages."
            )

        if args.target == ProjectBuildTargets.MAIN_TARGET:
            targets = list(project.working_tree.sources)
        else:
            targets = [args.target]

        build_tree = project.working_tree.clone()
        for target in targets:
            build_tree.build_targets.add_filter_stage(target, expr=filter_expr, params=filter_args)

    if args.apply:
        log.info("Filtering...")

        if is_target and not args.dst_dir:
            for target in targets:
                dataset = project.working_tree.make_dataset(build_tree.make_pipeline(target))

                # Source might be missing in the working dir, so we specify
                # the output directory.
                # We specify save_media here as a heuristic. It can probably
                # be improved by checking if there are images in the dataset
                # directory.
                dataset.save(project.source_data_dir(target), save_media=True)

            log.info("Finished")
        else:
            dataset, _project = parse_full_revpath(args.target, project)
            if _project:
                scope_add(_project)

            dst_dir = args.dst_dir or dataset.data_path
            if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
                raise CliException(
                    "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
                )
            dst_dir = osp.abspath(dst_dir)

            dataset.filter(filter_expr, *filter_args)
            dataset.save(dst_dir, save_media=True)

            log.info("Results have been saved to '%s'" % dst_dir)

    if is_target and args.stage:
        project.working_tree.config.update(build_tree.config)
        project.working_tree.save()

    return 0



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import inspect
import os
import os.path as osp
import shutil
from typing import TYPE_CHECKING, Callable, Dict, Iterable, Optional, Type, Union, overload

from datumaro.cli.util.compare import DistanceCompareVisualizer
from datumaro.components.comparator import DistanceComparator, EqualityComparator, TableComparator
from datumaro.components.dataset import Dataset, IDataset
from datumaro.components.environment import Environment
from datumaro.components.errors import DatasetError
from datumaro.components.exporter import Exporter
from datumaro.components.filter import (
    UserFunctionAnnotationsFilter,
    UserFunctionDatasetFilter,
    XPathAnnotationsFilter,
    XPathDatasetFilter,
)
from datumaro.components.launcher import Launcher
from datumaro.components.merge import DEFAULT_MERGE_POLICY, get_merger
from datumaro.components.transformer import ModelTransform, Transform
from datumaro.components.validator import TaskType, Validator
from datumaro.util import parse_str_enum_value
from datumaro.util.scope import on_error_do, scoped

if TYPE_CHECKING:
    from datumaro.components.annotation import Annotation
    from datumaro.components.dataset_base import DatasetItem

__all__ = ["HLOps"]


class HLOps:
    """High-level dataset operations for Python API."""

    @staticmethod
    def compare(
        first_dataset: IDataset,
        second_dataset: IDataset,
        report_dir: Optional[str] = None,
        method: str = "table",
        **kwargs,
    ) -> IDataset:
        """
        Compare two datasets and optionally save a comparison report.

        Args:
            first_dataset (IDataset): The first dataset to compare.
            second_dataset (IDataset): The second dataset to compare.
            report_dir (Optional[str], optional): The directory path to save the comparison report. Defaults to None.
            method (str, optional): The comparison method to use. Possible values are "table", "equality", "distance".
                Defaults to "table".
            **kwargs: Additional keyword arguments that can be passed to the comparison method.

        Returns:
            IDataset: The result of the comparison.

        Raises:
            ValueError: If the method is "distance" and report_dir is not specified.

        Example:
            comparator = Comparator()
            result = comparator.compare(first_dataset, second_dataset, report_dir="./comparison_report")
            print(result)
        """
        if method == "table":
            comparator = TableComparator()
            h_table, m_table, l_table, result_dict = comparator.compare_datasets(
                first_dataset, second_dataset
            )
            if report_dir:
                comparator.save_compare_report(h_table, m_table, l_table, result_dict, report_dir)

        elif method == "equality":
            comparator = EqualityComparator(**kwargs)
            output = comparator.compare_datasets(first_dataset, second_dataset)
            if report_dir:
                comparator.save_compare_report(output, report_dir)

        elif method == "distance":
            if not report_dir:
                raise ValueError(
                    "Please specify report_dir to save comparision result for DistanceComparator."
                )
            output_format = kwargs.pop("output_format", "simple")
            comparator = DistanceComparator(**kwargs)
            with DistanceCompareVisualizer(
                save_dir=report_dir,
                comparator=comparator,
                output_format=output_format,
            ) as visualizer:
                visualizer.save(first_dataset, second_dataset)

        return 0

    @staticmethod
    def transform(
        dataset: IDataset,
        method: Union[str, Type[Transform]],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> IDataset:
        """
        Applies some function to dataset items.

        Results are computed lazily, if the transform supports this.

        Args:
            dataset: The dataset to be transformed
            method: The transformation to be applied to the dataset.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the environment
                set by the 'env' argument
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the transformation

        Returns: a wrapper around the input dataset
        """

        if isinstance(method, str):
            if env is None:
                env = Environment()
            method = env.transforms[method]

        if not (inspect.isclass(method) and issubclass(method, Transform)):
            raise TypeError(f"Unexpected 'method' argument type: {type(method)}")

        produced = method(dataset, **kwargs)

        return Dataset(source=produced, env=env)

    @overload
    @staticmethod
    def filter(
        dataset: IDataset,
        expr: str,
        *,  # pylint: disable=redefined-builtin
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> IDataset:
        """
        Filters out some dataset items or annotations, using a custom filter
        expression.

        Args:
            dataset: The dataset to be filtered
            expr: XPath-formatted filter expression
                (e.g. `/item[subset = 'train']`, `/item/annotation[label = 'cat']`)
            filter_annotations: Indicates if the filter should be
                applied to items or annotations
            remove_empty: When filtering annotations, allows to
                exclude empty items from the resulting dataset

        Returns: a wrapper around the input dataset, which is computed lazily
            during iteration
        """
        ...

    @overload
    @staticmethod
    def filter(
        dataset: IDataset,
        filter_func: Union[
            Callable[[DatasetItem], bool], Callable[[DatasetItem, Annotation], bool]
        ],
        *,  # pylint: disable=redefined-builtin
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> IDataset:
        """
        Filters out some dataset items or annotations, using a user-provided filter
        Python function.

        Results are stored in-place. Modifications are applied lazily.

        Args:
            filter_func: User-provided Python function for filtering
            filter_annotations: Indicates if the filter should be
                applied to items or annotations
            remove_empty: When filtering annotations, allows to
                exclude empty items from the resulting dataset

        Returns: a wrapper around the input dataset, which is computed lazily
            during iteration

        Example:
            - (`filter_annotations=False`) This is an example of filtering
                dataset items with images larger than 1024 pixels::

                from datumaro.components.media import Image

                def filter_func(item: DatasetItem) -> bool:
                    h, w = item.media_as(Image).size
                    return h > 1024 or w > 1024

                filtered = HLOps.filter(
                    dataset=dataset,
                    filter_func=filter_func,
                    filter_annotations=False,
                )
                # No items with an image height or width greater than 1024
                filtered_items = [item for item in filtered]

            - (`filter_annotations=True`) This is an example of removing bounding boxes
                sized greater than 50% of the image size::

                from datumaro.components.media import Image
                from datumaro.components.annotation import Annotation, Bbox

                def filter_func(item: DatasetItem, ann: Annotation) -> bool:
                    # If the annotation is not a Bbox, do not filter
                    if not isinstance(ann, Bbox):
                        return False

                    h, w = item.media_as(Image).size
                    image_size = h * w
                    bbox_size = ann.h * ann.w

                    # Accept Bboxes smaller than 50% of the image size
                    return bbox_size < 0.5 * image_size

                filtered = HLOps.filter(
                    dataset=dataset,
                    filter_func=filter_func,
                    filter_annotations=True,
                )
                # No bounding boxes with a size greater than 50% of their image
                filtered_items = [item for item in filtered]
        """

    def filter(
        dataset: IDataset,
        expr_or_filter_func: Union[
            str, Callable[[DatasetItem], bool], Callable[[DatasetItem, Annotation], bool]
        ],
        *,  # pylint: disable=redefined-builtin
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ):
        if isinstance(expr_or_filter_func, str):
            expr = expr_or_filter_func
            return (
                HLOps.transform(
                    dataset, XPathAnnotationsFilter, xpath=expr, remove_empty=remove_empty
                )
                if filter_annotations
                else HLOps.transform(dataset, XPathDatasetFilter, xpath=expr)
            )
        elif callable(expr_or_filter_func):
            filter_func = expr_or_filter_func
            return (
                HLOps.transform(
                    dataset,
                    UserFunctionAnnotationsFilter,
                    filter_func=filter_func,
                    remove_empty=remove_empty,
                )
                if filter_annotations
                else HLOps.transform(dataset, UserFunctionDatasetFilter, filter_func=filter_func)
            )
        raise TypeError(expr_or_filter_func)

    @staticmethod
    def merge(
        *datasets: Dataset,
        merge_policy: str = DEFAULT_MERGE_POLICY,
        report_path: Optional[str] = None,
        **kwargs,
    ) -> Dataset:
        """
        Merge `datasets` according to `merge_policy`. You have to choose an appropriate `merge_policy`
        for your purpose. The available merge policies are "union", "intersect", and "exact".
        For more details about the merge policies, please refer to :func:`get_merger`.
        """

        merger = get_merger(merge_policy, **kwargs)
        merged = merger(*datasets)
        env = Environment.merge(
            [
                dataset.env
                for dataset in datasets
                if hasattr(dataset, "env")
                # TODO: Sometimes, there is dataset which is not exactly "Dataset",
                # e.g., VocClassificationBase. this should be fixed and every object from
                # Dataset.import_from should have "Dataset" type.
            ]
        )
        if report_path:
            merger.save_merge_report(report_path)
        return Dataset(source=merged, env=env)

    @staticmethod
    def run_model(
        dataset: IDataset,
        model: Union[Launcher, Type[ModelTransform]],
        *,
        batch_size: int = 1,
        append_annotation: bool = False,
        num_workers: int = 0,
        **kwargs,
    ) -> IDataset:
        """
        Run the model on the dataset item media entities, such as images,
        to obtain pseudo labels and add them as dataset annotations.

        Args:
            dataset: The dataset to be transformed
            model: The model to be applied to the dataset
            batch_size: The number of dataset items processed
                simultaneously by the model
            append_annotation: Whether append new annotation to existed annotations
            num_workers: The number of worker threads to use for parallel inference.
                Set to 0 for single-process mode. Default is 0.
            **kwargs: Parameters for the model

        Returns: a wrapper around the input dataset, which is computed lazily
            during iteration
        """

        if isinstance(model, Launcher):
            return HLOps.transform(
                dataset,
                ModelTransform,
                launcher=model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        elif inspect.isclass(model) and issubclass(model, ModelTransform):
            return HLOps.transform(
                dataset,
                model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        else:
            raise TypeError(f"Unexpected model argument type: {type(model)}")

    @staticmethod
    @scoped
    def export(
        dataset: IDataset,
        path: str,
        format: Union[str, Type[Exporter]],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> None:
        """
        Saves the input dataset in some format.

        Args:
            dataset: The dataset to be saved
            path: The output directory
            format: The desired output format for the dataset.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the environment set by the 'env' argument
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the export format
        """

        if isinstance(format, str):
            if env is None:
                env = Environment()
            exporter = env.exporters[format]
        else:
            exporter = format

        if not (inspect.isclass(exporter) and issubclass(exporter, Exporter)):
            raise TypeError(f"Unexpected 'format' argument type: {type(exporter)}")

        path = osp.abspath(path)
        if not osp.exists(path):
            on_error_do(shutil.rmtree, path, ignore_errors=True)
        os.makedirs(path, exist_ok=True)

        exporter.convert(dataset, save_dir=path, **kwargs)

    @staticmethod
    def validate(
        dataset: IDataset,
        task: Union[str, TaskType],
        *,
        env: Optional[Environment] = None,
        **kwargs,
    ) -> Dict:
        """
        Checks dataset annotations for correctness relatively to a task type.

        Args:
            dataset: The dataset to check
            task: Target task type - classification, detection etc.
            env: A plugin collection. If not set, the built-in plugins are used
            **kwargs: Parameters for the validator

        Returns: a dictionary with validation results
        """

        task = parse_str_enum_value(task, TaskType).name

        if env is None:
            env = Environment()

        validator: Validator = env.validators[task](**kwargs)
        return validator.validate(dataset)

    @staticmethod
    def aggregate(dataset: Dataset, from_subsets: Iterable[str], to_subset: str) -> Dataset:
        subset_names = set(dataset.subsets().keys())

        for subset in from_subsets:
            if subset not in subset_names:
                raise DatasetError(
                    f"{subset} is not found in the subset names ({subset_names}) in the dataset."
                )

        return HLOps.transform(
            dataset, "map_subsets", mapping={subset: to_subset for subset in from_subsets}
        )



# Copyright (C) 2023-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT
from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Set, Tuple

import numpy as np

from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import (
    AttributeDefinedButNotFound,
    BrokenAnnotation,
    DatasetValidationError,
    EmptyCaption,
    EmptyLabel,
    FarFromAttrMean,
    FarFromCaptionMean,
    FarFromLabelMean,
    FewSamplesInAttribute,
    FewSamplesInLabel,
    ImbalancedAttribute,
    ImbalancedCaptions,
    ImbalancedDistInAttribute,
    ImbalancedDistInCaption,
    ImbalancedDistInLabel,
    ImbalancedLabels,
    InvalidValue,
    LabelDefinedButNotFound,
    MissingAnnotation,
    MissingAttribute,
    MissingLabelCategories,
    MultiLabelAnnotations,
    NegativeLength,
    OnlyOneAttributeValue,
    OnlyOneLabel,
    OutlierInCaption,
    UndefinedAttribute,
    UndefinedLabel,
)
from datumaro.components.validator import Severity, TaskType, Validator


@dataclass
class StatsData:
    categories: Dict[str, Dict[str, Set[str]]]
    undefined_attribute: Set[Tuple[str, str, str]]  # item_id, item_subset, label_name, attr_name
    undefined_label: Set[Tuple[str, str, str]]  # item_id, item_subset, label_name
    missing_attribute: Set[Tuple[str, str, str, str]]  # item_id, item_subset, label_name, attr_name
    missing_label: Set[Tuple[str, str]]  # item_id, item_subset


@dataclass
class ClsStatsData(StatsData):
    multiple_label: Set[Tuple[str, str]]  # item_id, item_subset


@dataclass
class DetStatsData(StatsData):
    invalid_value: Set[Tuple[str, str]]  # item_id, item_subset
    negative_length: Set[Tuple[str, str]]  # item_id, item_subset


@dataclass
class SegStatsData(StatsData):
    invalid_value: Set[Tuple[str, str]]  # item_id, item_subset


@dataclass
class TblStatsData:
    categories: Dict[str, Dict[str, Set[str]]]
    empty_label: Set[Tuple[str, str]]  # item_id, item_subset
    empty_caption: Set[Tuple[str, str]]  # item_id, item_subset
    missing_annotations: Set[Tuple[str, str]]  # item_id, item_subset
    broken_annotations: Set[Tuple[str, str]]  # item_id, item_subset


class _BaseAnnStats:
    ATTR_WARNINGS = {
        AttributeDefinedButNotFound,
        FewSamplesInAttribute,
        ImbalancedAttribute,
        MissingAttribute,
        OnlyOneAttributeValue,
        UndefinedAttribute,
    }

    def __init__(
        self,
        label_categories: LabelCategories,
        warnings: List[DatasetValidationError],
        few_samples_thr: None,
        imbalance_ratio_thr: None,
        far_from_mean_thr: None,
        dominance_thr: None,
        topk_bins_ratio: None,
    ):
        self.label_categories = label_categories
        self.warnings = set(warnings)

        self.few_samples_thr = few_samples_thr
        self.imbalance_ratio_thr = imbalance_ratio_thr
        self.far_from_mean_thr = far_from_mean_thr
        self.dominance_thr = dominance_thr
        self.topk_bins_ratio = topk_bins_ratio

        self.stats = StatsData(
            categories={},
            undefined_attribute=set(),
            undefined_label=set(),
            missing_attribute=set(),
            missing_label=set(),
        )

        for label_cat in self.label_categories.items:
            label_name = label_cat.name
            self.stats.categories[label_name] = {
                "cnt": 0,
                "type": set(),
                "attributes": {attr: {} for attr in label_cat.attributes},
            }

        # Create a dictionary that maps warning types to their corresponding update functions
        self.update_item_functions = set()
        self.update_ann_functions = set()
        self.update_report_functions = set()

        if LabelDefinedButNotFound in self.warnings:
            self.update_ann_functions.add(self._update_label_stats)
            self.update_report_functions.add(self._check_label_defined_but_not_found)
        if FewSamplesInLabel in self.warnings:
            self.update_ann_functions.add(self._update_label_stats)
            self.update_report_functions.add(self._check_few_samples_in_label)
        if ImbalancedLabels in self.warnings:
            self.update_ann_functions.add(self._update_label_stats)
            self.update_report_functions.add(self._check_imbalanced_labels)
        if OnlyOneLabel in self.warnings:
            self.update_ann_functions.add(self._update_label_stats)
            self.update_report_functions.add(self._check_only_one_label)
        if AttributeDefinedButNotFound in self.warnings:
            self.update_ann_functions.add(self._update_attribute_stats)
            self.update_report_functions.add(self._check_attribute_defined_but_not_found)
        if FewSamplesInAttribute in self.warnings:
            self.update_ann_functions.add(self._update_attribute_stats)
            self.update_report_functions.add(self._check_few_samples_in_attribute)
        if ImbalancedAttribute in self.warnings:
            self.update_ann_functions.add(self._update_attribute_stats)
            self.update_report_functions.add(self._check_imbalanced_attribute)
        if OnlyOneAttributeValue in self.warnings:
            self.update_ann_functions.add(self._update_attribute_stats)
            self.update_report_functions.add(self._check_only_one_attribute)
        if MissingAnnotation in self.warnings:
            self.update_item_functions.add(self._update_missing_annotation)
            self.update_report_functions.add(self._check_missing_label)
        if UndefinedLabel in self.warnings:
            self.update_ann_functions.add(self._update_undefined_label)
            self.update_report_functions.add(self._check_undefined_label)
        if MissingAttribute in self.warnings:
            self.update_ann_functions.add(self._update_missing_attribute)
            self.update_report_functions.add(self._check_missing_attribute)
        if UndefinedAttribute in self.warnings:
            self.update_ann_functions.add(self._update_undefined_attribute)
            self.update_report_functions.add(self._check_undefined_attribute)

    def _update_label_stats(self, item_key, annotation: Annotation):
        if annotation.label in self.label_categories:
            label_name = self.label_categories[annotation.label].name
            self.stats.categories[label_name]["cnt"] += 1
            self.stats.categories[label_name]["type"].add(annotation.type)

    def _update_attribute_stats(self, item_key, annotation):
        if annotation.label in self.label_categories:
            label_name = self.label_categories[annotation.label].name
            for attr, value in annotation.attributes.items():
                if attr in self.stats.categories[label_name]["attributes"]:
                    attr_cnt = self.stats.categories[label_name]["attributes"][attr].get(
                        str(value), 0
                    )
                    attr_cnt += 1

    def _update_missing_annotation(self, item):
        item_key = (item.id, item.subset)
        if sum([1 for ann in item.annotations]) == 0:
            self.stats.missing_label.add(item_key)

    def _update_undefined_label(self, item_key, annotation):
        if annotation.label not in self.label_categories:
            self.stats.undefined_label.add(item_key + (str(annotation.label),))

    def _update_missing_attribute(self, item_key, annotation):
        if annotation.label in self.label_categories:
            label_name = self.label_categories[annotation.label].name
            for attr in self.stats.categories[label_name]["attributes"]:
                if attr not in annotation.attributes:
                    self.stats.missing_attribute.add(item_key + (label_name, attr))

    def _update_undefined_attribute(self, item_key, annotation):
        if annotation.label in self.label_categories:
            label_name = self.label_categories[annotation.label].name
            for attr in annotation.attributes:
                if self.stats.categories[label_name]["attributes"].get(attr, None):
                    continue
                self.stats.undefined_attribute.add(item_key + (label_name, attr))
        else:
            for attr in annotation.attributes:
                self.stats.undefined_attribute.add(item_key + (str(annotation.label), attr))

    def _generate_validation_report(self, error, *args, **kwargs):
        return [error(*args, **kwargs)]

    def _check_missing_label_categories(self, stats: StatsData):
        validation_reports = []

        if len(stats.categories) == 0:
            validation_reports += self._generate_validation_report(
                MissingLabelCategories, Severity.error
            )

        return validation_reports

    def _check_missing_label(self, stats: StatsData):
        validation_reports = []

        items_missing_label = stats.missing_label
        for item_id, item_subset in items_missing_label:
            validation_reports += self._generate_validation_report(
                MissingAnnotation, Severity.warning, item_id, item_subset, "label"
            )

        return validation_reports

    def _check_missing_attribute(self, stats: StatsData):
        validation_reports = []

        items_missing_attr = stats.missing_attribute
        for item_id, item_subset, label_name, attr_name in items_missing_attr:
            details = (item_subset, label_name, attr_name)
            validation_reports += self._generate_validation_report(
                MissingAttribute, Severity.warning, item_id, *details
            )

        return validation_reports

    def _check_undefined_label(self, stats: StatsData):
        validation_reports = []

        items_undefined_label = stats.undefined_label
        for item_id, item_subset, label_name in items_undefined_label:
            details = (item_subset, label_name)
            validation_reports += self._generate_validation_report(
                UndefinedLabel, Severity.error, item_id, *details
            )

        return validation_reports

    def _check_undefined_attribute(self, stats: StatsData):
        validation_reports = []

        items_undefined_attr = stats.undefined_attribute
        for item_id, item_subset, label_name, attr_name in items_undefined_attr:
            details = (item_subset, label_name, attr_name)
            validation_reports += self._generate_validation_report(
                UndefinedAttribute, Severity.error, item_id, *details
            )

        return validation_reports

    def _check_label_defined_but_not_found(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            if stats.categories[label_name]["cnt"] == 0:
                validation_reports += self._generate_validation_report(
                    LabelDefinedButNotFound, Severity.warning, label_name
                )

        return validation_reports

    def _check_attribute_defined_but_not_found(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            for attr_name, dist in stats.categories[label_name]["attributes"].items():
                if dist:
                    continue
                details = (label_name, attr_name)
                validation_reports += self._generate_validation_report(
                    AttributeDefinedButNotFound, Severity.warning, *details
                )

        return validation_reports

    def _check_only_one_label(self, stats: StatsData):
        labels_found = []
        for label_name in stats.categories:
            if stats.categories[label_name]["cnt"] > 0:
                labels_found.append(label_name)

        validation_reports = []
        if len(labels_found) == 1:
            validation_reports += self._generate_validation_report(
                OnlyOneLabel, Severity.info, labels_found[0]
            )

        return validation_reports

    def _check_only_one_attribute(self, stats: StatsData):
        validation_reports = []
        for label_name in stats.categories:
            for attr_name, values in stats.categories[label_name]["attributes"].items():
                for attr_value, cnt in values.items():
                    if cnt != 1:
                        continue
                    details = (label_name, attr_name, attr_value)
                    validation_reports += self._generate_validation_report(
                        OnlyOneAttributeValue, Severity.info, *details
                    )

        return validation_reports

    def _check_few_samples_in_label(self, stats: StatsData):
        validation_reports = []
        for label_name in stats.categories:
            if stats.categories[label_name]["cnt"] < self.few_samples_thr:
                validation_reports += self._generate_validation_report(
                    FewSamplesInLabel,
                    Severity.info,
                    label_name,
                    stats.categories[label_name]["cnt"],
                )

        return validation_reports

    def _check_few_samples_in_attribute(self, stats: StatsData):
        validation_reports = []
        for label_name in stats.categories:
            for attr_name, values in stats.categories[label_name]["attributes"].items():
                for attr_value, cnt in values.items():
                    if cnt >= self.few_samples_thr:
                        continue
                    details = (label_name, attr_name, attr_value, cnt)
                    validation_reports += self._generate_validation_report(
                        FewSamplesInAttribute, Severity.info, *details
                    )

        return validation_reports

    def _check_imbalanced_labels(self, stats: StatsData):
        validation_reports = []

        counts = [stats.categories[label_name]["cnt"] for label_name in stats.categories]

        if len(counts) == 0:
            return validation_reports

        count_max = np.max(counts)
        count_min = np.min(counts)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= self.imbalance_ratio_thr:
            validation_reports += self._generate_validation_report(ImbalancedLabels, Severity.info)

        return validation_reports

    def _check_imbalanced_attribute(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            for attr_name, attr_vals in stats.categories[label_name]["attributes"].items():
                counts = [cnt for cnt in attr_vals.values()]

                if len(counts) == 0:
                    continue

                count_max = np.max(counts)
                count_min = np.min(counts)
                balance = count_max / count_min if count_min > 0 else float("inf")
                if balance >= self.imbalance_ratio_thr:
                    validation_reports += self._generate_validation_report(
                        ImbalancedAttribute, Severity.info, label_name, attr_name
                    )

        return validation_reports

    def update_item(self, item: DatasetItem):
        for func in self.update_item_functions:
            func(item)

    def update_ann(self, item_key, annotation: Annotation):
        for func in self.update_ann_functions:
            func(item_key, annotation)

    def generate_report(self):
        reports = []
        for func in self.update_report_functions:
            report = func(self.stats)
            if report:
                reports += report
        return reports


class ClsStats(_BaseAnnStats):
    def __init__(
        self,
        label_categories: LabelCategories,
        warnings: set,
        few_samples_thr: None,
        imbalance_ratio_thr: None,
        far_from_mean_thr: None,
        dominance_thr: None,
        topk_bins_ratio: None,
    ):
        super().__init__(
            label_categories=label_categories,
            warnings=warnings,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_thr=dominance_thr,
            topk_bins_ratio=topk_bins_ratio,
        )

        self.stats = ClsStatsData(
            categories={},
            undefined_attribute=set(),
            undefined_label=set(),
            missing_attribute=set(),
            missing_label=set(),
            multiple_label=set(),
        )

        for label_cat in self.label_categories.items:
            label_name = label_cat.name
            self.stats.categories[label_name] = {
                "cnt": 0,
                "type": set(),
                "attributes": {attr: {} for attr in label_cat.attributes},
            }

        if MultiLabelAnnotations in self.warnings:
            self.update_item_functions.add(self._update_multi_label)
            self.update_report_functions.add(self._check_multiple_label)

    def _update_multi_label(self, item: DatasetItem):
        item_key = (item.id, item.subset)
        if sum([1 for ann in item.annotations if ann.type == AnnotationType.label]) > 1:
            self.stats.multiple_label.add(item_key)

    def _check_multiple_label(self, stats: ClsStatsData):
        validation_reports = []

        items_multiple_label = stats.multiple_label
        for item_id, item_subset in items_multiple_label:
            validation_reports += self._generate_validation_report(
                MultiLabelAnnotations, Severity.error, item_id, item_subset
            )

        return validation_reports


class DetStats(_BaseAnnStats):
    def __init__(
        self,
        label_categories: LabelCategories,
        warnings: set,
        few_samples_thr: None,
        imbalance_ratio_thr: None,
        far_from_mean_thr: None,
        dominance_thr: None,
        topk_bins_ratio: None,
    ):
        super().__init__(
            label_categories=label_categories,
            warnings=warnings,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_thr=dominance_thr,
            topk_bins_ratio=topk_bins_ratio,
        )

        self.stats = DetStatsData(
            categories={},
            undefined_attribute=set(),
            undefined_label=set(),
            missing_attribute=set(),
            missing_label=set(),
            invalid_value=set(),
            negative_length=set(),
        )

        for label_cat in self.label_categories.items:
            label_name = label_cat.name
            self.stats.categories[label_name] = {
                "cnt": 0,
                "type": set(),
                "attributes": {attr: {} for attr in label_cat.attributes},
                "pts": [],  # (id, subset, ann_id, w, h, a, r)
            }

        if FarFromLabelMean in self.warnings:
            self.update_ann_functions.add(self._update_bbox_stats)
            self.update_report_functions.add(self._check_far_from_mean)
        if ImbalancedDistInLabel in self.warnings:
            self.update_ann_functions.add(self._update_bbox_stats)
            self.update_report_functions.add(self._check_imbalanced_dist_in_label)
        if FarFromAttrMean in self.warnings:
            self.update_ann_functions.add(self._update_bbox_stats)
            self.update_report_functions.add(self._check_far_from_mean)
        if ImbalancedDistInAttribute in self.warnings:
            self.update_ann_functions.add(self._update_bbox_stats)
            self.update_report_functions.add(self._check_imbalanced_dist_in_label)
        if InvalidValue in self.warnings:
            self.update_ann_functions.add(self._update_invalid_value)
            self.update_report_functions.add(self._check_invalid_value)
        if NegativeLength in self.warnings:
            self.update_ann_functions.add(self._update_negative_length)
            self.update_report_functions.add(self._check_negative_length)

    def _update_invalid_value(self, item_key: tuple, annotation: Annotation):
        _, _, w, h = annotation.get_bbox()

        if w == float("inf") or np.isnan(w):
            self.stats.invalid_value.add(item_key + (annotation.id, "width", w))

        if h == float("inf") or np.isnan(h):
            self.stats.invalid_value.add(item_key + (annotation.id, "height", h))

    def _update_negative_length(self, item_key: tuple, annotation: Annotation):
        _, _, w, h = annotation.get_bbox()

        if w < 1:
            self.stats.negative_length.add(item_key + (annotation.id, "width", w))

        if h < 1:
            self.stats.negative_length.add(item_key + (annotation.id, "height", h))

    def _update_bbox_stats(self, item_key: tuple, annotation: Annotation):
        if annotation.label not in self.label_categories:
            return

        _, _, w, h = annotation.get_bbox()
        area = annotation.get_area()
        if h != 0 and h != float("inf"):
            ratio = w / h
        else:
            ratio = float("nan")

        if not (
            w == float("inf")
            or np.isnan(w)
            or w < 1
            or h == float("inf")
            or np.isnan(h)
            or h < 1
            or np.isnan(ratio)
        ):
            label_name = self.label_categories[annotation.label].name
            self.stats.categories[label_name]["pts"].append(
                item_key + (annotation.id, w, h, area, ratio)
            )

    def _check_far_from_mean(self, stats: StatsData):
        def _far_from_mean(val, mean, stdev):
            thr = self.far_from_mean_thr
            return val > mean + (thr * stdev) or val < mean - (thr * stdev)

        validation_reports = []

        for label_name in stats.categories:
            prop = {"width": [], "height": [], "ratio": [], "area": []}
            for _, _, _, w, h, a, r in stats.categories[label_name]["pts"]:
                prop["width"].append(w)
                prop["height"].append(h)
                prop["ratio"].append(r)
                prop["area"].append(a)

            prop_stats = {}
            for p, vals in prop.items():
                prop_stats[p] = {}
                prop_stats[p]["mean"] = np.mean(vals)
                prop_stats[p]["stdev"] = np.std(vals)

            for item_id, item_subset, ann_id, w, h, a, r in stats.categories[label_name]["pts"]:
                item_prop = {"width": w, "height": h, "ratio": r, "area": a}

                for p in prop_stats.keys():
                    if _far_from_mean(item_prop[p], prop_stats[p]["mean"], prop_stats[p]["stdev"]):
                        details = (
                            item_subset,
                            label_name,
                            ann_id,
                            f"bbox {p}",
                            prop_stats[p]["mean"],
                            item_prop[p],
                        )
                        validation_reports += self._generate_validation_report(
                            FarFromLabelMean, Severity.warning, item_id, *details
                        )

        return validation_reports

    def _check_imbalanced_dist_in_label(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            prop = {"width": [], "height": [], "ratio": [], "area": []}
            for _, _, _, w, h, a, r in stats.categories[label_name]["pts"]:
                prop["width"].append(w)
                prop["height"].append(h)
                prop["area"].append(a)
                prop["ratio"].append(r)

            for p, vals in prop.items():
                counts, _ = np.histogram(vals)

                n_bucket = len(counts)
                if n_bucket < 2:
                    continue

                topk = max(1, int(np.around(n_bucket * self.topk_bins_ratio)))

                topk_values = np.sort(counts)[-topk:]
                ratio = np.sum(topk_values) / np.sum(counts)
                if ratio >= self.dominance_thr:
                    details = (label_name, f"bbox {p}")
                    validation_reports += self._generate_validation_report(
                        ImbalancedDistInLabel, Severity.info, *details
                    )

        return validation_reports

    def _check_negative_length(self, stats: DetStatsData):
        validation_reports = []

        items_neg_len = stats.negative_length
        for item_id, item_subset, ann_id, prop, val in items_neg_len:
            details = (item_subset, ann_id, f"bbox {prop}", val)
            validation_reports += self._generate_validation_report(
                NegativeLength, Severity.error, item_id, *details
            )

        return validation_reports

    def _check_invalid_value(self, stats: DetStatsData):
        validation_reports = []

        items_invalid_val = stats.invalid_value
        for item_id, item_subset, ann_id, prop, val in items_invalid_val:
            details = (item_subset, ann_id, f"bbox {prop}")
            validation_reports += self._generate_validation_report(
                InvalidValue, Severity.error, item_id, *details
            )

        return validation_reports


class SegStats(_BaseAnnStats):
    def __init__(
        self,
        label_categories: LabelCategories,
        warnings: set,
        few_samples_thr: None,
        imbalance_ratio_thr: None,
        far_from_mean_thr: None,
        dominance_thr: None,
        topk_bins_ratio: None,
    ):
        super().__init__(
            label_categories=label_categories,
            warnings=warnings,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_thr=dominance_thr,
            topk_bins_ratio=topk_bins_ratio,
        )

        self.stats = SegStatsData(
            categories={},
            undefined_attribute=set(),
            undefined_label=set(),
            missing_attribute=set(),
            missing_label=set(),
            invalid_value=set(),
        )

        for label_cat in self.label_categories.items:
            label_name = label_cat.name
            self.stats.categories[label_name] = {
                "cnt": 0,
                "type": set(),
                "attributes": {attr: {} for attr in label_cat.attributes},
                "pts": [],  # (id, subset, ann_id, w, h, a, r)
            }

        if FarFromLabelMean in self.warnings:
            self.update_ann_functions.add(self._update_mask_stats)
            self.update_report_functions.add(self._check_far_from_mean)
        if ImbalancedDistInLabel in self.warnings:
            self.update_ann_functions.add(self._update_mask_stats)
            self.update_report_functions.add(self._check_imbalanced_dist_in_label)
        if FarFromAttrMean in self.warnings:
            self.update_ann_functions.add(self._update_mask_stats)
            self.update_report_functions.add(self._check_far_from_mean)
        if ImbalancedDistInAttribute in self.warnings:
            self.update_ann_functions.add(self._update_mask_stats)
            self.update_report_functions.add(self._check_imbalanced_dist_in_label)
        if InvalidValue in self.warnings:
            self.update_ann_functions.add(self._update_invalid_value)
            self.update_report_functions.add(self._check_invalid_value)

    def _update_invalid_value(self, item_key: tuple, annotation: Annotation):
        _, _, w, h = annotation.get_bbox()

        if w == float("inf") or np.isnan(w):
            self.stats.invalid_value.add(item_key + (annotation.id, "width", w))

        if h == float("inf") or np.isnan(h):
            self.stats.invalid_value.add(item_key + (annotation.id, "height", h))

    def _update_mask_stats(self, item_key: tuple, annotation: Annotation):
        if annotation.label not in self.label_categories:
            return

        _, _, w, h = annotation.get_bbox()
        area = annotation.get_area()

        if not (w == float("inf") or np.isnan(w) or h == float("inf") or np.isnan(h)):
            label_name = self.label_categories[annotation.label].name
            self.stats.categories[label_name]["pts"].append(
                item_key + (annotation.id, w, h, area, 1)
            )

    def _check_far_from_mean(self, stats: StatsData):
        def _far_from_mean(val, mean, stdev):
            thr = self.far_from_mean_thr
            return val > mean + (thr * stdev) or val < mean - (thr * stdev)

        validation_reports = []

        for label_name in stats.categories:
            prop = {"width": [], "height": [], "area": []}
            for _, _, _, w, h, a, _ in stats.categories[label_name]["pts"]:
                prop["width"].append(w)
                prop["height"].append(h)
                prop["area"].append(a)

            prop_stats = {}
            for p, vals in prop.items():
                prop_stats[p] = {}
                prop_stats[p]["mean"] = np.mean(vals)
                prop_stats[p]["stdev"] = np.std(vals)

            for item_id, item_subset, ann_id, w, h, a, _ in stats.categories[label_name]["pts"]:
                item_prop = {"width": w, "height": h, "area": a}

                for p in prop_stats.keys():
                    if _far_from_mean(item_prop[p], prop_stats[p]["mean"], prop_stats[p]["stdev"]):
                        details = (
                            item_subset,
                            label_name,
                            ann_id,
                            f"polygon/mask/ellipse {p}",
                            prop_stats[p]["mean"],
                            item_prop[p],
                        )
                        validation_reports += self._generate_validation_report(
                            FarFromLabelMean, Severity.warning, item_id, *details
                        )

        return validation_reports

    def _check_imbalanced_dist_in_label(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            prop = {"width": [], "height": [], "area": []}
            for _, _, _, w, h, a, _ in stats.categories[label_name]["pts"]:
                prop["width"].append(w)
                prop["height"].append(h)
                prop["area"].append(a)

            for p, vals in prop.items():
                counts, _ = np.histogram(vals)

                n_bucket = len(counts)
                if n_bucket < 2:
                    continue

                topk = max(1, int(np.around(n_bucket * self.topk_bins_ratio)))

                topk_values = np.sort(counts)[-topk:]
                ratio = np.sum(topk_values) / np.sum(counts)
                if ratio >= self.dominance_thr:
                    details = (label_name, f"polygon/mask/ellipse {p}")
                    validation_reports += self._generate_validation_report(
                        ImbalancedDistInLabel, Severity.info, *details
                    )

        return validation_reports

    def _check_invalid_value(self, stats: DetStatsData):
        validation_reports = []

        items_invalid_val = stats.invalid_value
        for item_id, item_subset, ann_id, prop, val in items_invalid_val:
            details = (item_subset, ann_id, f"polygon/mask/ellipse {prop}")
            validation_reports += self._generate_validation_report(
                InvalidValue, Severity.error, item_id, *details
            )

        return validation_reports


from pandas.api.types import CategoricalDtype


class TblStats(_BaseAnnStats):
    def __init__(
        self,
        categories: dict,
        warnings: set,
        few_samples_thr: None,
        imbalance_ratio_thr: None,
        far_from_mean_thr: None,
        dominance_thr: None,
        topk_bins_ratio: None,
    ):
        self.caption_categories = categories.get(AnnotationType.caption, [])
        self.label_categories = categories.get(AnnotationType.label, [])

        self.warnings = set(warnings)

        self.few_samples_thr = few_samples_thr
        self.imbalance_ratio_thr = imbalance_ratio_thr
        self.far_from_mean_thr = far_from_mean_thr
        self.dominance_thr = dominance_thr
        self.topk_bins_ratio = topk_bins_ratio

        self.stats = TblStatsData(
            categories={},
            empty_label=set(),
            empty_caption=set(),
            missing_annotations=set(),
            broken_annotations=set(),
        )

        all_categories = [(cat.name, cat.dtype) for cat in self.caption_categories] + [
            (label_group.name, CategoricalDtype)
            for label_group in self.label_categories.label_groups
        ]
        for name, dtype in all_categories:
            self.stats.categories[name] = {
                "cnt": 0,
                "type": dtype,
                "ann_type": set(),
                "caption": [],
            }
        self.categories_len = len(all_categories)

        self.caption_columns = [cat.name for cat in self.caption_categories]
        self.label_columns = [
            label_group.name for label_group in self.label_categories.label_groups
        ]

        # Create a dictionary that maps warning types to their corresponding update functions
        self.update_item_functions = set()
        self.update_ann_functions = set()
        self.update_report_functions = set()

        if MissingAnnotation in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_item_functions.add(self._update_missing_annotation)
            self.update_report_functions.add(self._check_missing_annotation)
        if BrokenAnnotation in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_item_functions.add(self._update_broken_annotation)
            self.update_report_functions.add(self._check_broken_annotation)

        if FewSamplesInLabel in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_few_samples_in_label)
        if ImbalancedLabels in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_imbalanced_labels)
        if EmptyLabel in self.warnings:
            self.update_item_functions.add(self._update_empty_label)
            self.update_report_functions.add(self._check_empty_label)

        if EmptyCaption in self.warnings:
            self.update_item_functions.add(self._update_empty_caption)
            self.update_report_functions.add(self._check_empty_caption)
        if ImbalancedCaptions in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_imbalanced_captions)
        if ImbalancedDistInCaption in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_imbalanced_dist_in_caption)
        if FarFromCaptionMean in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_far_from_caption_mean)
        if OutlierInCaption in self.warnings:
            self.update_ann_functions.add(self._update_annotation_stats)
            self.update_report_functions.add(self._check_outlier_in_caption)

    def _update_missing_annotation(self, item):
        item_key = (item.id, item.subset)
        if len(item.annotations) == 0:
            self.stats.missing_annotations.add(item_key)

    def _update_broken_annotation(self, item):
        item_key = (item.id, item.subset)
        if len(item.annotations) < self.categories_len:
            self.stats.broken_annotations.add(item_key)

    def _update_empty_label(self, item):
        if len(item.annotations) < self.categories_len:
            annotation_check = deepcopy(self.label_columns)
            for ann in item.annotations:
                if ann.type == AnnotationType.label:
                    label = self.label_categories[ann.label].name.split(":")[0]
                    annotation_check.remove(label)
            for ann in annotation_check:
                item_key = (item.id, item.subset, ann)
                self.stats.empty_label.add(item_key)

    def _update_empty_caption(self, item):
        if len(item.annotations) < self.categories_len:
            annotation_check = deepcopy(self.caption_columns)
            for ann in item.annotations:
                if ann.type == AnnotationType.caption:
                    label = next(
                        (cat for cat in self.caption_columns if ann.caption.startswith(cat)), None
                    )
                    annotation_check.remove(label)
            for ann in annotation_check:
                item_key = (item.id, item.subset, ann)
                self.stats.empty_caption.add(item_key)

    def _update_annotation_stats(self, item_key, annotation: Annotation):
        if annotation.type == AnnotationType.label:
            label_name = self.label_categories[annotation.label].name.split(":")[0]
            self.stats.categories[label_name]["cnt"] += 1
            self.stats.categories[label_name]["ann_type"].add(annotation.type)
        elif annotation.type == AnnotationType.caption:
            for cat in self.caption_columns:
                if annotation.caption.startswith(cat):
                    self.stats.categories[cat]["cnt"] += 1
                    self.stats.categories[cat]["ann_type"].add(annotation.type)
                    caption = annotation.caption.split(cat + ":")[-1]
                    self.stats.categories[cat]["caption"].append(
                        item_key + (annotation.id, caption)
                    )
                    break

    def _check_missing_annotation(self, stats: StatsData):
        validation_reports = []

        items_missing_annotations = stats.missing_annotations
        for item_id, item_subset in items_missing_annotations:
            validation_reports += self._generate_validation_report(
                MissingAnnotation, Severity.warning, item_id, item_subset, "label or caption"
            )

        return validation_reports

    def _check_broken_annotation(self, stats: StatsData):
        validation_reports = []

        items_broken_annotations = stats.broken_annotations
        for item_id, item_subset in items_broken_annotations:
            validation_reports += self._generate_validation_report(
                BrokenAnnotation, Severity.warning, item_id, item_subset, "label or caption"
            )

        return validation_reports

    def _check_empty_label(self, stats: StatsData):
        validation_reports = []

        items_empty_label = stats.empty_label
        for item_id, item_subset, col in items_empty_label:
            validation_reports += self._generate_validation_report(
                EmptyLabel, Severity.warning, item_id, item_subset, col
            )

        return validation_reports

    def _check_empty_caption(self, stats: StatsData):
        validation_reports = []

        items_empty_caption = stats.empty_caption
        for item_id, item_subset, col in items_empty_caption:
            validation_reports += self._generate_validation_report(
                EmptyCaption, Severity.warning, item_id, item_subset, col
            )

        return validation_reports

    def _check_imbalanced_labels(self, stats: StatsData):
        validation_reports = []

        counts = [
            stats.categories[label_name]["cnt"]
            for label_name in stats.categories.keys()
            if list(stats.categories[label_name]["ann_type"])[0] == AnnotationType.label
        ]

        if len(counts) == 0:
            return validation_reports

        count_max = np.max(counts)
        count_min = np.min(counts)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= self.imbalance_ratio_thr:
            validation_reports += self._generate_validation_report(ImbalancedLabels, Severity.info)

        return validation_reports

    def _check_imbalanced_captions(self, stats: StatsData):
        validation_reports = []

        counts = [
            stats.categories[label_name]["cnt"]
            for label_name in stats.categories.keys()
            if list(stats.categories[label_name]["ann_type"])[0] == AnnotationType.caption
        ]

        if len(counts) == 0:
            return validation_reports

        count_max = np.max(counts)
        count_min = np.min(counts)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= self.imbalance_ratio_thr:
            validation_reports += self._generate_validation_report(
                ImbalancedCaptions, Severity.info
            )

        return validation_reports

    def _check_far_from_caption_mean(self, stats: StatsData):
        def _far_from_mean(val, mean, stdev):
            thr = self.far_from_mean_thr
            return val > mean + (thr * stdev) or val < mean - (thr * stdev)

        validation_reports = []

        for label_name in stats.categories:
            type_ = stats.categories[label_name]["type"]
            if type_ in [float, int]:
                captions = [
                    type_(caption[3]) for caption in stats.categories[label_name]["caption"]
                ]
                prop_stats = {}
                prop_stats["mean"] = np.mean(captions)
                prop_stats["stdev"] = np.std(captions)

                upper_bound = prop_stats["mean"] + (self.far_from_mean_thr * prop_stats["stdev"])
                lower_bound = prop_stats["mean"] - (self.far_from_mean_thr * prop_stats["stdev"])
                for item_id, item_subset, ann_id, caption in stats.categories[label_name][
                    "caption"
                ]:
                    if _far_from_mean(type_(caption), prop_stats["mean"], prop_stats["stdev"]):
                        details = (
                            item_subset,
                            label_name,
                            prop_stats["mean"],
                            upper_bound,
                            lower_bound,
                            type_(caption),
                        )
                        validation_reports += self._generate_validation_report(
                            FarFromCaptionMean, Severity.info, item_id, *details
                        )

        return validation_reports

    def _check_imbalanced_dist_in_caption(self, stats: StatsData):
        validation_reports = []

        for label_name in stats.categories:
            type_ = stats.categories[label_name]["type"]
            if type_ in [float, int]:
                captions = [
                    type_(caption[3]) for caption in stats.categories[label_name]["caption"]
                ]
                counts, _ = np.histogram(captions)

                n_bucket = len(counts)
                if n_bucket < 2:
                    continue

                topk = max(1, int(np.around(n_bucket * self.topk_bins_ratio)))

                topk_values = np.sort(counts)[-topk:]
                ratio = np.sum(topk_values) / np.sum(counts)
                if ratio >= self.dominance_thr:
                    validation_reports += self._generate_validation_report(
                        ImbalancedDistInCaption, Severity.info, label_name
                    )
        return validation_reports

    def _check_outlier_in_caption(self, stats: StatsData):
        validation_reports = []

        for label_name, category in stats.categories.items():
            type_ = category["type"]
            if type_ in [float, int]:
                captions = np.array([type_(caption[3]) for caption in category["caption"]])

                if captions.size == 0:
                    continue

                # Calculate Q1 (25th percentile) and Q3 (75th percentile)
                Q1 = np.quantile(captions, 0.25)
                Q3 = np.quantile(captions, 0.75)
                IQR = Q3 - Q1

                # Calculate the acceptable rangepsrc/datumaro/plugins/validators.py
                lower_bound = Q1 - 1.5 * IQR
                upper_bound = Q3 + 1.5 * IQR

                for item_id, item_subset, ann_id, caption in category["caption"]:
                    val = type_(caption)
                    if (val < lower_bound) | (val > upper_bound):
                        details = (
                            item_subset,
                            label_name,
                            upper_bound,
                            lower_bound,
                            val,
                        )
                        validation_reports += self._generate_validation_report(
                            OutlierInCaption, Severity.info, item_id, *details
                        )

        return validation_reports


class ConfigurableValidator(Validator, CliPlugin):
    DEFAULT_FEW_SAMPLES_THR = 1
    DEFAULT_IMBALANCE_RATIO_THR = 50
    DEFAULT_FAR_FROM_MEAN_THR = 5
    DEFAULT_DOMINANCE_RATIO_THR = 0.8
    DEFAULT_TOPK_BINS = 0.1

    ALL_WARNINGS = {
        AttributeDefinedButNotFound,  # annotation level
        FarFromAttrMean,  # annotation level
        FarFromLabelMean,  # annotation level
        FewSamplesInAttribute,  # annotation level
        FewSamplesInLabel,  # annotation level
        ImbalancedAttribute,  # annotation level
        ImbalancedDistInAttribute,  # annotation level: bbox
        ImbalancedDistInLabel,  # annotation level: bbox
        ImbalancedLabels,  # annotation level
        InvalidValue,  # annotation level
        LabelDefinedButNotFound,  # item level
        MissingAnnotation,  # item level
        MissingAttribute,  # annotation level
        MissingLabelCategories,  # dataset level
        MultiLabelAnnotations,  # item level
        NegativeLength,  # annotation level
        OnlyOneAttributeValue,  # annotation level
        OnlyOneLabel,  # annotation level
        UndefinedAttribute,  # annotation level
        UndefinedLabel,  # annotation level
    }

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-fs",
            "--few-samples-thr",
            default=cls.DEFAULT_FEW_SAMPLES_THR,
            type=int,
            help="Threshold for giving a warning for minimum number of "
            "samples per class (default: %(default)s)",
        )
        parser.add_argument(
            "-ir",
            "--imbalance-ratio-thr",
            default=cls.DEFAULT_IMBALANCE_RATIO_THR,
            type=int,
            help="Threshold for giving data imbalance warning. "
            "IR(imbalance ratio) = majority/minority "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-m",
            "--far-from-mean-thr",
            default=cls.DEFAULT_FAR_FROM_MEAN_THR,
            type=float,
            help="Threshold for giving a warning that data is far from mean. "
            "A constant used to define mean +/- k * standard deviation "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-dr",
            "--dominance-ratio-thr",
            default=cls.DEFAULT_DOMINANCE_RATIO_THR,
            type=float,
            help="Threshold for giving a warning for bounding box imbalance. "
            "Dominace_ratio = ratio of Top-k bin to total in histogram "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-k",
            "--topk-bins",
            default=cls.DEFAULT_TOPK_BINS,
            type=float,
            help="Ratio of bins with the highest number of data "
            "to total bins in the histogram. A value in the range [0, 1] "
            "(default: %(default)s)",
        )
        return parser

    def __init__(
        self,
        tasks: List[TaskType] = [
            TaskType.classification,
            TaskType.detection,
            TaskType.segmentation,
            TaskType.tabular,
        ],
        warnings: Set[DatasetValidationError] = ALL_WARNINGS,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        self.tasks = tasks
        self.warnings = warnings

        self.all_stats = {task: None for task in tasks}

        self.few_samples_thr = few_samples_thr if few_samples_thr else self.DEFAULT_FEW_SAMPLES_THR
        self.imbalance_ratio_thr = (
            imbalance_ratio_thr if imbalance_ratio_thr else self.DEFAULT_IMBALANCE_RATIO_THR
        )
        self.far_from_mean_thr = (
            far_from_mean_thr if far_from_mean_thr else self.DEFAULT_FAR_FROM_MEAN_THR
        )
        self.dominance_thr = (
            dominance_ratio_thr if dominance_ratio_thr else self.DEFAULT_DOMINANCE_RATIO_THR
        )
        self.topk_bins_ratio = topk_bins if topk_bins else self.DEFAULT_TOPK_BINS

    def _init_stats_collector(self, categories):
        for task in self.tasks:
            kwargs = {
                "few_samples_thr": self.few_samples_thr,
                "imbalance_ratio_thr": self.imbalance_ratio_thr,
                "far_from_mean_thr": self.far_from_mean_thr,
                "dominance_thr": self.dominance_thr,
                "topk_bins_ratio": self.topk_bins_ratio,
            }
            if task == TaskType.classification:
                self.all_stats[task] = ClsStats(
                    label_categories=categories, warnings=self.warnings, **kwargs
                )
            elif task == TaskType.detection:
                self.all_stats[task] = DetStats(
                    label_categories=categories, warnings=self.warnings, **kwargs
                )
            elif task == TaskType.segmentation:
                self.all_stats[task] = SegStats(
                    label_categories=categories, warnings=self.warnings, **kwargs
                )
            elif task == TaskType.tabular:
                self.all_stats[task] = TblStats(
                    categories=categories, warnings=self.warnings, **kwargs
                )

    def _get_stats_collector(self, ann_type, task_type):
        if task_type == [TaskType.tabular]:
            return self.all_stats.get(TaskType.tabular, None)
        else:
            if ann_type == AnnotationType.label:
                return self.all_stats.get(TaskType.classification, None)
            elif ann_type == AnnotationType.bbox:
                return self.all_stats.get(TaskType.detection, None)
            elif ann_type in [
                AnnotationType.mask,
                AnnotationType.polygon,
                AnnotationType.ellipse,
            ]:
                return self.all_stats.get(TaskType.segmentation, None)
            else:
                return None

    def compute_statistics(self, dataset):
        if self.tasks == [TaskType.tabular]:
            categories_input = dataset.categories()
        else:
            categories_input = dataset.categories()[AnnotationType.label]
        self._init_stats_collector(categories=categories_input)

        for item in dataset:
            for stats_collector in self.all_stats.values():
                stats_collector.update_item(item)

            item_key = (item.id, item.subset)
            for annotation in item.annotations:
                stats_collector = self._get_stats_collector(
                    ann_type=annotation.type, task_type=self.tasks
                )
                if stats_collector:
                    stats_collector.update_ann(item_key, annotation)

        return {task: stats_collector for task, stats_collector in self.all_stats.items()}

    def generate_reports(self, task_stats):
        reports = {}
        for task, stats_collector in task_stats.items():
            reports[task] = stats_collector.generate_report()

        return reports



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Sequence

import numpy as np
from tokenizers import Tokenizer

from datumaro.components.annotation import Annotation, HashKey
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import MediaTypeError
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.launcher import OpenvinoLauncher


class ExplorerLauncher(OpenvinoLauncher):
    def __init__(
        self,
        description=None,
        weights=None,
        interpreter=None,
        model_dir=None,
        model_name=None,
        output_layers=None,
        device=None,
    ):
        super().__init__(
            description, weights, interpreter, model_dir, model_name, output_layers, device
        )

        self._device = device or "cpu"
        self._output_blobs = next(iter(self._net.outputs))
        self._tokenizer = None

    def _tokenize(self, texts: str, context_length: int = 77, truncate: bool = True):
        if not self._tokenizer:
            checkpoint = "openai/clip-vit-base-patch32"
            self._tokenizer = Tokenizer.from_pretrained(checkpoint)
        tokens = self._tokenizer.encode(texts).ids
        result = np.zeros((1, context_length))

        if len(tokens) > context_length:
            if truncate:
                eot_token = tokens.ids[-1]
                tokens = tokens[:context_length]
                tokens[-1] = eot_token

        for i, token in enumerate(tokens):
            result[:, i] = token
        return result

    def _compute_hash(self, features):
        features = np.sign(features)
        hash_key = np.clip(features, 0, None)
        hash_key = hash_key.astype(np.uint8)
        hash_key = np.packbits(hash_key, axis=-1)
        return hash_key

    def infer_text(self, text: str, use_prompt: bool = True) -> HashKey:
        prompt_text = f"a photo of a {text}" if use_prompt else text
        inputs = self._tokenize(prompt_text)
        preds = self.infer(inputs)
        anns = self.postprocess(preds[0], None)
        return anns[0]

    def infer_item(self, item: DatasetItem) -> HashKey:
        anns = self.launch([item])[0]
        return anns[0]

    def launch(self, batch: Sequence[DatasetItem]) -> List[List[Annotation]]:
        outputs = super().launch(batch)
        return outputs

    def type_check(self, item):
        if not isinstance(item.media, Image):
            raise MediaTypeError(f"Media type should be Image, Current type={type(item.media)}")
        return True



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from . import modification, versioning


def get_project_commands():
    return [
        ("Project modification:", None, ""),
        ("add", modification.add, "Add dataset"),  # TODO: We will deprecated it with import soon.
        ("create", modification.create, "Create empty project"),
        ("export", modification.export, "Export dataset in some format"),
        ("import", modification.import_, "Import dataset"),
        ("remove", modification.remove, "Remove dataset"),
        ("", None, ""),
        ("Project versioning:", None, ""),
        ("checkout", versioning.checkout, "Switch to another branch or revision"),
        ("commit", versioning.commit, "Commit changes in tracked files"),
        ("log", versioning.log, "List history"),
        ("info", versioning.info, "Print project information"),
        ("status", versioning.status, "Display current branch and revision status"),
    ]



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import os.path as osp
from collections import OrderedDict

import numpy as np

from datumaro.components.annotation import AnnotationType, HashKey
from datumaro.util import dump_json_file, find, parse_json_file

DATASET_META_FILE = "dataset_meta.json"
DATASET_HASHKEY_FILE = "hash_keys.json"
DATASET_HASHKEY_FOLDER = "hash_key_meta"


def is_meta_file(path):
    return osp.splitext(osp.basename(path))[1] == ".json"


def has_meta_file(path):
    return osp.isfile(get_meta_file(path))


def has_hashkey_file(path):
    return osp.isfile(get_hashkey_file(path))


def get_meta_file(path):
    return osp.join(path, DATASET_META_FILE)


def get_hashkey_file(path):
    hashkey_folder_path = osp.join(path, DATASET_HASHKEY_FOLDER)
    return osp.join(hashkey_folder_path, DATASET_HASHKEY_FILE)


def parse_meta_file(path):
    meta_file = path
    if osp.isdir(path):
        meta_file = get_meta_file(path)

    dataset_meta = parse_json_file(meta_file)

    label_map = OrderedDict()

    for label in dataset_meta.get("labels", []):
        label_map[label] = None

    colors = dataset_meta.get("segmentation_colors", [])
    for i, label in dataset_meta.get("label_map", {}).items():
        label_map[label] = None

        if any(colors) and colors[int(i)] is not None:
            label_map[label] = tuple(colors[int(i)])

    return label_map


def save_meta_file(path, categories):
    dataset_meta = {}

    labels = [label.name for label in categories[AnnotationType.label]]
    dataset_meta["labels"] = labels

    if categories.get(AnnotationType.mask):
        label_map = {}
        segmentation_colors = []
        for i, color in categories[AnnotationType.mask].colormap.items():
            if color:
                segmentation_colors.append([int(color[0]), int(color[1]), int(color[2])])
                label_map[str(i)] = labels[i]
        dataset_meta["label_map"] = label_map
        dataset_meta["segmentation_colors"] = segmentation_colors

        bg_label = find(
            categories[AnnotationType.mask].colormap.items(), lambda x: x[1] == (0, 0, 0)
        )
        if bg_label is not None:
            dataset_meta["background_label"] = str(bg_label[0])

    meta_file = path
    if osp.isdir(path):
        meta_file = get_meta_file(path)

    dump_json_file(meta_file, dataset_meta, indent=True)


def parse_hashkey_file(path):
    meta_file = path
    if osp.isdir(path):
        meta_file = get_hashkey_file(path)

    if not osp.exists(meta_file):
        return None

    dataset_meta = parse_json_file(meta_file)

    hashkey_dict = OrderedDict()
    for id_, hashkey in dataset_meta.get("hashkey", {}).items():
        hashkey_dict[id_] = hashkey

    return hashkey_dict


def save_hashkey_file(path, item_list):
    dataset_hashkey = {}

    if osp.isdir(path):
        meta_file = get_hashkey_file(path)
    hashkey_folder_path = osp.join(path, DATASET_HASHKEY_FOLDER)
    if not osp.exists(hashkey_folder_path):
        os.makedirs(hashkey_folder_path)

    hashkey_dict = parse_hashkey_file(path)
    if not hashkey_dict:
        hashkey_dict = {}

    for item in item_list:
        item_id = item.id
        item_subset = item.subset
        for annotation in item.annotations:
            if isinstance(annotation, HashKey):
                hashkey = annotation.hash_key
                break
        hashkey_dict.update({item_subset + "/" + item_id: hashkey.tolist()})

    dataset_hashkey["hashkey"] = hashkey_dict

    dump_json_file(meta_file, dataset_hashkey, indent=True)


def load_hash_key(path, dataset):
    if not os.path.isdir(path) or not has_hashkey_file(path):
        return dataset

    hashkey_dict = parse_hashkey_file(path)
    for item in dataset:
        hash_key = hashkey_dict[item.subset + "/" + item.id]
        item.annotations.append(HashKey(hash_key=np.asarray(hash_key, dtype=np.uint8)))
    return dataset



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict

import pyarrow as pa

from datumaro.components.dataset_base import DatasetItem
from datumaro.plugins.data_formats.datumaro_binary.mapper.annotation import AnnotationListMapper
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper, Mapper

from .media import MediaMapper


class DatasetItemMapper(Mapper):
    @staticmethod
    def forward(obj: DatasetItem, **options) -> Dict[str, Any]:
        return {
            "id": obj.id,
            "subset": obj.subset,
            "media": MediaMapper.forward(obj.media, **options.get("media", {})),
            "annotations": AnnotationListMapper.forward(obj.annotations),
            "attributes": DictMapper.forward(obj.attributes),
        }

    @staticmethod
    def backward(idx: int, table: pa.Table, table_path: str) -> DatasetItem:
        return DatasetItem(
            id=table.column("id")[idx].as_py(),
            subset=table.column("subset")[idx].as_py(),
            media=MediaMapper.backward(
                media_struct=table.column("media")[idx],
                idx=idx,
                table=table,
                table_path=table_path,
            ),
            annotations=AnnotationListMapper.backward(table.column("annotations")[idx].as_py())[0],
            attributes=DictMapper.backward(table.column("attributes")[idx].as_py())[0],
        )
